/*

   Lehrveranstaltung Computergraphik,
   OpenGL Praktikum
   "Hello World"

Autor: Prof. Dr. Ralf Dörner, HS RheinMain
Version 1.0

*/

// für Textausgaben auf der Konsole:
#include <stdio.h>

// hier die notwendigen Includes - es genügt glut.h einzubinden, da
// diese wiederum GL/gl.h und alle weiteren benötigten Include-
// Dateien einbindet
#include "util.h"
#include "custom_figures.h"
#include "swarm.h"
#include <GL/glut.h>
#include <vector>
#include <ctime>
#include <string>
#include <memory>
#include <cmath>

#define PYRAMIDS 0
#define FISH 1
#define SWARM_SIZE 100

using namespace std;

static shared_ptr<Pyramid> p1, p2, p3;
static shared_ptr<Swarm> swarm;
static shared_ptr<Shark> shark;
static shared_ptr<FishTank> tank;
static shared_ptr<Plant> plant1;
static int shading = FLAT_SHADING;
static int scene = PYRAMIDS;
static int lightIntensity = 50;
static int dimDir = 1;
static Vector tankCenter;
static int moves = 0;
static bool fishrotate = true;
static bool followShark = false;
static bool sharkEyes = false;
static time_t lastSec;
static int lastX, lastY;
static bool rotate;
static bool zoom;

static float theta, phi;
static float eyeDistance;
static int fps = 0;

static Vector focus (0, 0, 0);
static Vector eye (0, 0, 0);

static bool lights;

void updateWindowTitle()
{
  glutSetWindowTitle((to_string(swarm->fish.size() + 1) +
    string("/400 Fischen im Aquarium @ ") +
    to_string(fps) + string(" FPS")).c_str());
  fps = 0;
}

shared_ptr<Pyramid> getP1()
{
  shared_ptr<Pyramid> result = make_shared<Pyramid>();
  result->scale(Vector(4, 6, 4));
  result->material.diffuse = Color(1, 0, 0);
  return result;
}

shared_ptr<Pyramid> getP2()
{
  shared_ptr<Pyramid> result = p1->clone();
  result->rotate(Vector(1, 0, 0), 180, Vector(0, 6, 0));
  result->material.diffuse = Color(0, 1, 0);
  return result;
}

shared_ptr<Pyramid> getP3()
{
  shared_ptr<Pyramid> result = p1->clone();
  result->scale(Vector(2, 2, 2));
  result->move(Vector(0, -12, 0));
  result->move(Vector(5, -3, 0));
  result->material.diffuse = Color(1, 0, 0);
  return result;
}

shared_ptr<Swarm> getSwarm()
{
  return make_shared<Swarm>(SWARM_SIZE);
}

shared_ptr<Fish> getFish()
{
  return make_shared<Fish>();
}

shared_ptr<FishTank> getTank(shared_ptr<Swarm> swarm, shared_ptr<Shark> shark)
{
  return make_shared<FishTank>(*swarm, *shark);
}

shared_ptr<Shark> getShark(shared_ptr<Swarm> swarm)
{
  shared_ptr<Shark> shark = make_shared<Shark>();
  shark->move(swarm->center());
  shark->applyTransformations();
  while(swarm->collides(*shark)) {
    shark->move(Vector(10, 0, 0));
    shark->applyTransformations();
  }
  return shark;
}

shared_ptr<Plant> getPlant(const char* filename, int width, int height)
{
  return make_shared<Plant>(filename, width, height);
}

void addToGlowingOnesEmissionBrightness(float amount)
{
  Color emission = swarm->fish[0].material.emission;
  emission.b += amount;
  if (emission.b > 1)
    emission.b = 1;
  if (emission.b < 0)
    emission.b = 0;
  swarm->fish[0].material.emission = emission;
}

void setEyePosition(float x, float y, float z)
{
  eye.x = x;
  eye.y = y;
  eye.z = z;
}

void setFocus(float x, float y, float z)
{
  focus.x = x;
  focus.y = y;
  focus.z = z;
}

void calcEyePosition()
{
  setEyePosition(
      focus.x + eyeDistance*cos(phi)*sin(theta),
      focus.y + eyeDistance*sin(phi)*sin(theta),
      focus.z + eyeDistance*cos(theta)
  );
}
// die Funktion display enthält alle Befehle, die zum Zeichnen des
// Bildes benötigt werden
void display( void )
{
  time_t now;
  time(&now);
  if (difftime(now, lastSec) >= 1) {
    updateWindowTitle();
    lastSec = now;
  } else {
    fps++;
  }

  // Zeigt von nach "vorne" zeigenden Polygonen nur die Kanten.
  // glPolygonMode(GL_FRONT, GL_LINE);
  // Zeigt von nach "vorne" _und_ nach hinten Zeigenden Polygonen nur die Kanten.
  // glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);

  // zunächst wird das bisher gezeichnete Bild gelöscht,
  // indem das Bild komplett in der Hintergrundfarbe gezeichnet
  // wird. Ausserdem wird die Verdeckungsrechnung initialisiert
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);


  // die folgenden drei Befehle dienen dazu anzugeben, von wo aus
  // die Szene betrachtet wird (d.h. unter welchem Blickwinkel
  // das Bild berechnet wird
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();
  // die ersten drei Argumente sind die Koordinaten, wo der
  // Betrachter steht, die nächsten drei Argumente geben die
  // Koordinaten an, wohin der Betrachter blickt und die letzten
  // drei Argumente geben an, wo oben ist (bei uns ist y-Richtung
  // oben

  gluLookAt(eye.x, eye.y, eye.z,
            focus.x, focus.y, focus.z,
            0.0, 1.0, 0.0);

  if (lights) {
    GLfloat ambientColor[] = {0.5, 0.5, 0.5, 1.0};
    glLightModelfv(GL_LIGHT_MODEL_AMBIENT, ambientColor);

    glEnable(GL_LIGHTING);
    glEnable(GL_LIGHT0);
    GLfloat spotLightColor0 [] = {
      lightIntensity / 100.0f,
      lightIntensity / 100.0f,
      lightIntensity / 100.0f, 1};
    GLfloat spotLightPos0 [] = {5.0, 20.0, 20.0, 1.0};
    glLightfv(GL_LIGHT0, GL_DIFFUSE, spotLightColor0);
    glLightfv(GL_LIGHT0, GL_POSITION, spotLightPos0);
    lightIntensity = (lightIntensity + 1 * dimDir) % 100;
    if (lightIntensity == 99 || lightIntensity == 50)
      dimDir = -dimDir;

  } else {
    glDisable(GL_LIGHTING);
  }

  if (scene == PYRAMIDS) {
    // die Zeichenfarbe wird auf rot gesetzt
    glColor3f(1.0, 0.0, 0.0);
    // P1
    p1->paint(shading);
    p1->rotate(Vector(0, 1, 0), 3, p1->center);
    // P3
    p3->paint(shading);
    p3->rotate(Vector(0, 1, 0), 2, p3->center);
    glColor3f(0.0, 1.0, 0.0);

    // P2
    p2->paint(shading);
    p2->rotate(Vector(0, 1, 0), -1, p2->center);
  } else {
    swarm->move(*tank, *shark);
    shark->turnTowards(swarm->center(), 4);
    shark->moveForward(4);
    swarm->paint(shading);
    shark->paint(shading);
    plant1->paint(shading);
    glEnable (GL_BLEND);
    glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    tank->paint(shading);
    if (followShark) {
      setEyePosition(shark->center.x - 1, 150, shark->center.z);
      setFocus(shark->center.x, shark->center.y, shark->center.z);
    } else if (sharkEyes) {
      Vector eyes = shark->center + (shark->direction * shark->boundingRadius);
      setEyePosition(eyes.x, eyes.y, eyes.z);
      Vector focus = eyes + shark->direction;
      setFocus(focus.x, focus.y, focus.z);
    }
  }

  // Sorge dafür, dass wirklich alle glBefehle abgearbeitet und
  // nicht gepuffert werden
  glFlush();

  // Vertausche die Backbuffer und Frontbuffer: das fertig
  // gezeichnete Bild soll jetzt angezeigt werden
  glutSwapBuffers();

  // Sorge dafür, dass das Bild im nächsten Schleifendurchlauf
  // der glutMainLoop wieder gezeichnet wird
  glutPostRedisplay();

} // end display


// die Funktion reshape enthält alle Befehle, die man benötigt,
// um auf Veränderungen der Fenstergröße zu reagieren
void reshape(int w, int h)
{

  // hier wird die Perspektive gesetzt
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  gluPerspective(60.0, w / h, 0.1, 300.0);
  // Durch einen kleineren Öffnungswinkel der Kamera wird die
  // Szene "herangezoomt"
  // gluPerspective(30.0, w / h, 0.1, 100.0);

  // Durch einen kleineren wert für far clipping plane wird die
  // Szene schneller ausgeblendet wenn von weitem auf die Szene
  // geschaut wird.
  // gluPerspective(60.0, w / h, 0.1, 34.0);
  glViewport(0.0, 0.0, w, h);
  glMatrixMode(GL_MODELVIEW);

} // end reshape


// die Funktion wird aufgerufen, sobald eine Taste der Tastur
// gedrückt wurde. Es steht die Info zur Verfügung, welche
// Taste gedrückt wurde und wo sich die Maus zu diesem Zeitpunkt
// befand
void keyboard(unsigned char key, int x, int y)
{
  switch (key) {
  case '1':
    setEyePosition(0, 0, 0);
    setFocus(5, -3, 0);
    break;
  case '2':
    setEyePosition(0, 50, 1);
    setFocus(0, 6, 0);
    break;
  case '3':
    setEyePosition(5, 20, 1);
    setFocus(5, -3, 0);
    break;
  case '4':
    setEyePosition(0, -20, 1);
    setFocus(0, 6, 0);
    break;
  case '5':
    setEyePosition(tankCenter.x - 1, 150, tankCenter.z);
    setFocus(tankCenter.x, tankCenter.y, tankCenter.z);
    break;
  case '6':
    setEyePosition(tankCenter.x, tankCenter.y, tankCenter.z + 80);
    setFocus(tankCenter.x, tankCenter.y, tankCenter.z);
    break;
  case '7':
    setEyePosition(tankCenter.x - 80, tankCenter.y, tankCenter.z);
    setFocus(tankCenter.x, tankCenter.y, tankCenter.z);
    break;
  case 'b':
    lights = !lights;
    break;
  case 'f':
    shading = shading == SMOOTH_SHADING ? FLAT_SHADING : SMOOTH_SHADING;
    break;
  case 'p':
    scene = scene == PYRAMIDS ? FISH : PYRAMIDS;
    break;
  case '+':
    addToGlowingOnesEmissionBrightness(0.05);
    break;
  case '-':
    addToGlowingOnesEmissionBrightness(-0.05);
  case 's':
    swarm->spawnFish();
  }
  if (key >= '0' && key <= '9') {
    followShark = false;
    sharkEyes = false;
    if (key == '8')
      followShark = true;
    else if (key == '9')
      sharkEyes = true;
  }
}


// die Funktion wird aufgerufen, sobald eine Taste der Maus
// gedrückt wurde. Es steht die Info zur Verfügung, welche
// Taste betroffen ist, ob sie gedrückt oder losgelassen wurde
// und wo sich die Maus zu diesem Zeitpunkt befand
void mouse(int button, int state, int x, int y)
{
  rotate = false;
  zoom = false;
  if (GLUT_DOWN == state) {
    if (GLUT_LEFT_BUTTON == button) {
        rotate = true;
    } else if (GLUT_RIGHT_BUTTON == button) {
      zoom = true;
    }
  }
  lastX = x;
  lastY = y;
}

void motion(int x, int y)
{
  if (rotate) {
    theta += (x - lastX) * 0.005;
    phi += (y - lastY) * 0.005;
    calcEyePosition();
  } else if (zoom) {
    eyeDistance += (y - lastY) * 0.1;
    calcEyePosition();
  }
  lastX = x;
  lastY = y;
}

// Hauptfunktion
int main(int argc, char** argv)
{
  p1 = getP1();
  p2 = getP2();
  p3 = getP3();
  swarm = getSwarm();
  shark = getShark(swarm);
  tank = getTank(swarm, shark);
  tankCenter = tank->center;
  plant1 = getPlant("pflanze1.bmp", 512, 512);
  theta = 0;
  phi = 0;
  eyeDistance = 20;
  lights = true;
  calcEyePosition();

  // Initialisiere die GLUT
  glutInit(&argc,argv);

  // Setze den Display Mode: Farben werden in RGB angegeben,
  // es wird eine Verdeckungsrechnung durchgeführt und mit
  // zwei Buffern gearbeitet
  glutInitDisplayMode (GLUT_RGB | GLUT_DEPTH | GLUT_DOUBLE);

  // Setze die Fenstergröße auf 500 x 500
  glutInitWindowSize(500,500);

  // Erzeuge ein Fenster auf dem Bildschirm an Position
  // 100, 0 (Bildschirmkoordinaten)
  glutInitWindowPosition(100,0);

  // Erzeuge das Fenster
  glutCreateWindow("");

  // Bestimme, welcher Callback zum Zeichnen aufgerufen wird
  glutDisplayFunc(display);

  // Bestimme, welcher Callback bei Größenänderung des Fensters
  // aufgerufen wird
  glutReshapeFunc(reshape);

  // Bestimme, welcher Callback bei Betätigung der Tastatur
  // aufgerufen wird
  glutKeyboardFunc(keyboard);

  // Bestimme, welcher Callback bei Betätigen einer Maustaste
  // aufgerufen wird
  glutMouseFunc(mouse);

  glutMotionFunc(motion);

  // die Hintergrundfarbe wird auf gelb gesetzt
  // glClearColor(1.0, 1.0, 0.0, 1.0);

  // die Initialisierung der Verdeckungsrechnung ist 1.0
  glClearDepth(1.0);

  // die Verdeckungsrechnung wird eingeschaltet
  glEnable(GL_DEPTH_TEST);

  // Übergebe die Kontrolle nun an GLUT, die in einer Endlos-
  // schleife die Callbacks aufruft
  glutMainLoop();

}// end main
